# @(#) $Id: Malaga.pm 1.4 Fri, 27 Jun 2008 15:37:29 +0200 mxp $

package Malaga;

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);

require Exporter;
require DynaLoader;
require AutoLoader;

@ISA = qw(Exporter AutoLoader DynaLoader);
# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.
@EXPORT = qw(MORPHOLOGY SYNTAX);
@EXPORT_OK = qw(
		init
		analyze
		term
		get_info
		set_option
		value2string
		value2pretty
		string2value
	       );
# $Format: "$VERSION = '$ProjectVersion$';"$
$VERSION = '5.9';

sub MORPHOLOGY{return 0};
sub SYNTAX{return 1};

bootstrap Malaga $VERSION;

# Preloaded methods go here.

sub value2string
{
  my $value  = shift;
  my $result = "";

  if(ref($value) eq "HASH")
  {
     $result .= hash2string($value);
  }
  elsif(ref($value) eq "ARRAY")
  {
     $result .= array2string($value);
  }
  elsif(!ref ($value))
  {
     if($value =~ /([0-9])\.([0-9]+)(e?)([0-9]*)/)
     {
	$result .= sprintf("%f", $value);
     }
     else
     {
	$result .= $value;
     }
  }
  return $result;
}

sub hash2string
{
   my $hash   = shift;
   my $result = "";

   $result .= "[";

   my @attributes = keys %$hash;
   my $counter     = 0;
   my $number      = $#attributes;

   foreach my $attribute (@attributes)
   {
      $result .= $attribute . ": ";
      $result .= value2string($hash->{$attribute});
      if($counter != $number)
      {
	 $result .= ", ";
	 $counter++;
      }
   }
   $result .= "]";
   return $result;
}

sub array2string
{
   my $array  = shift;
   my $result = "";

   $result .= "<";
   my $counter = 0;

   foreach my $element (@$array)
   {
      $result .= value2string($element);
      if($counter != $#{$array})
      {
	 $result .= ", ";
	 $counter++;
      }
   }
   $result .= ">";
   return $result;
}

# Pretty printing

sub printtab 
{
   my $tab = shift;
   return " " x $tab;
}

sub value2pretty
{
   my $value = shift;
   my $tab;
   $tab = shift or $tab = 0;
   my $result = "";

   if(ref($value) eq "HASH")
   {
      $result .= hash2pretty($value, $tab);
   }
   elsif(ref($value) eq "ARRAY")
   {
      $result .= array2pretty($value, $tab);
   }
   elsif(!ref ($value))
   {
      if($value =~ /([0-9])\.([0-9]+)(e?)([0-9]*)/)
      {
	 $result .= sprintf("%f", $value);
      }
      else
      {
	 $result .= $value;
      }
   }
   return $result;
}

sub hash2pretty
{
   my $hash   = shift;
   my $tab    = shift;
   my $result = "";

   $result .= "[";

   my @attributes = keys %$hash;
   my $counter     = 0;
   my $number      = $#attributes;

   foreach my $attribute (@attributes)
   {
      my $attlength = length($attribute);
      my $newtab = $tab + $attlength + 3;
      $result .= $attribute . ": ";
      $result .= value2pretty($hash->{$attribute}, $newtab);
      if($counter != $number)
      {
	 $result .= ",\n";
	 $result .= printtab($tab+1);
	 $counter++;
      }
   }
   $result .= "]";
   return $result;
}

sub array2pretty
{
   my $array  = shift;
   my $tab    = shift;
   my $result = "";

   $result .= "<";
   my $counter = 0;

   foreach my $element (@$array)
   {
      $result .= value2pretty($element, $tab+1);
      if($counter != $#{$array})
      {
	 $result .= ",\n";
	 $result .= printtab($tab + 1);
	 $counter++;
      }
   }
   $result .= ">";
   return $result;
}

# Autoload methods go after =cut, and are processed by the autosplit program.

1;
__END__
# Below is the documentation for the module.

=head1 NAME

Malaga - Perl interface to Malaga grammars

=head1 SYNOPSIS

  use Malaga;
  Malaga::init("malaga_project_file.pro");
  $copyright = Malaga::get_info();
  Malaga::set_option("option string");

  @results = Malaga::analyze("string");

  print Malaga::value2string(\@results);
  print Malaga::value2pretty(\@results);

  $value = Malaga::string2value($string);

  Malaga::term;

=head1 DESCRIPTION

The Malaga extension for Perl enables you to have text linguistically
analyzed by Malaga grammars directly from your Perl program.

First, you have to load the desired Malaga grammar by means of init().
init() takes one argument, the name and path of a Malaga project file,
complete with the F<.pro> extension.  You can then use the get_info(),
set_option(), and analyze() functions.

get_info() returns a string with the informative message specified in
the C<info> lines of the project file.

set_option() allows to set options for the loaded grammar.  This
allows to override options specified in the project file.  The
possible options are described in the Malaga manual.

analyze() is used to analyze strings with the selected grammar until
you close the grammar with term().

analyze() takes one, two, or three arguments:

=over 4

=item *

The first argument is the string to be analyzed.  If no further
arguments are specified, it is analyzed morphologically.  This
argument is obligatory.  All other arguments are optional.  If an
optional argument is specified, all optional arguments before it must
also be specified, i.e., the second argument must be specified if the
third is given etc.

=item *

The optional second argument can be one of the constants C<MORPHOLOGY>
or C<SYNTAX>, indicating the desired type of analysis.

=item *

If the optional third argument is true, Malaga strings will be
enclosed in quotation marks.  If this option is used, the conversion
of a Malaga symbol will differ from the conversion of a Malaga string.
The former will result in a Perl string containing the name of the
symbol, while the latter will result in a Perl string containing the
value of the string surrounded by quotes.

=back

analyze() returns an array of references. Each element of the array
represents one analysis of the input (to be more precise: it is a
reference to a data structure containing the category of a final
state). If there is just a single analysis for a given input, you get
an array with one element; if it is ambigous, you get more
elements. If there is no analysis, you have got an empty list,
i.e. the first element of the list is not defined. The elements of the
array--if there are any--are always references. The type of the
references and the data they point to is grammar-dependent.

Malaga data types are mapped to Perl data types in the following ways:

 Malaga type           | Perl type
 ----------------------+------------------------
 Symbol                | string (i.e., a scalar)
 String                | string (i.e., a scalar)
 Floating point number | number (i.e., a scalar)
 List                  | list
 Record                | hash

So, the following Malaga structure

 [Form: [POS: Conjunction,
         Syn: [ConjType: Coord],
         Mor: [Weight: 1,
	       Position: free,
	       Structure: <[POS: Conjunction,
			   Type: WordStart,
			   Morpheme: "aber",
			   Allomorph: "aber"]>
	      ]
	],
  Sem: [BaseForm: "aber"]]

becomes the following in Perl :

 {Form => {POS => "Conjunction",
           Syn => {ConjType => "Coord"},
           Mor => {Weight => 1,
                   Position => "free",
                   Structure =>
                    [{POS => "Conjunction",
                      Type => "WordStart",
                      Morpheme => "aber",
                      Allomorph => "aber"}]
                  }
          },
  Sem  => {BaseForm => "aber"}}

Assuming that this is the first analysis in C<@results>, which was set
by C<@results = analyze("aber")>, you could access the value of the
"path" (in Malaga-speak) C<Form.Syn.ConjType> like this:

 $results[0]{Form}{Syn}{ConjType}

See below for another example; see perlref(1), perllol(1), and
perldsc(1) for more information on references and complex data
structures in Perl.

The argument to value2string() is a reference to a result structure as
returned by analyze(), or a part of such a structure, which is
converted to a string in Malaga notation.  Note that

  @results = analyze("word");
  print value2string(@results);

prints only the first analysis result, whereas

  @results = analyze("word");
  print value2string(\@results);

prints all analysis results.

value2pretty() works the same way as value2string() but returns a
``prettyprinted'' string, i.e., with elements indented according to
their nesting level.

string2value() reads a string as produced by value2string() or
value2pretty() and converts it into the corresponding Perl structure
(as returned by analyze()).  Note that init() must have been called
before calling string2value(), as it relies on a libmalaga function,
and that strings must be quoted.

=head1 EXAMPLE

 # Load the Malaga module and have
 # some functions exported
 use Malaga qw(set_option analyze get_info);

 # Initialize a grammar
 Malaga::init("/usr/local/lib/malaga/grammar.pro");

 # Print information about the grammar
 print get_info();

 # Analyze a word form morphologically
 @results = analyze("Meister");

 # Print a value of the result category
 print $results[0]{'Form'}{'POS'} . "\n";

 # Print the number of results
 print "There are ", scalar(@results),
       " analyses of 'Meister'\n";

 # Switch off morphologic filtering
 set_option("mfil off");

 # Analyze the same word form again
 @results = analyze("Meister");

 # See if you get a different number of results
 print "Now there are ", scalar(@results),
       " analyses of 'Meister'\n";

 # Terminate the grammar
 Malaga::term;

=head1 ERRORS

All functions except value2string() and value2pretty() throw an
exception if an error occurs.  Exceptions can be caught using I<eval>;
$@ contains the error message provided by libmalaga.

=head1 RESTRICTIONS

The following restrictions apply:

=over 4

=item *

You can only have one active grammar at a time.  This is a restriction
of Malaga.

=item *

There is currently no support for analysis trees.

=item *

The Malaga I<sort-records> option has no effect because records are
converted into hashes, in which elements are unordered.

=back

=head1 NOTES

The grammars in the examples are completely fictitious; actually
available grammars are likely to use different feature-value
structures.

=head1 VERSION

$ProjectVersion: 5.9 $

=head1 COMPATIBILITY

This version of the Malaga module is compatible with the libmalaga
version 6 interface (Malaga version 6.8) and with libmalaga versions 7
and 8.  Starting with version 7, libmalaga requires all input and
output to be encoded in UTF-8.  Thus all strings to be analyzed must
be encoded in UTF-8, and all strings returned by analyses will be
encoded in UTF-8 and are marked as such (i.e., utf8::is_utf8() is
true).  Refert to perluniintro(1) for more information about Perl's
Unicode handling.

=head1 AUTHOR

Michael Piotrowski <mxp@dynalabs.de>

=head1 SEE ALSO

perl(1), perlref(1), perllol(1), perldsc(1), and the Malaga
documentation.

=cut

